/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *         "The Daily Plan-It"  A Smart Date-Parsing Filter            *
 *                          Release 1.0                                *
 *                                                                     *
 *    Copyright (c) 1986 by David W. Burleigh. All rights reserved.    *
 *                                                                     *
 *    Usage:  plan-it [m/d[/y]] <filename                              *
 *                                                                     *
 *       or:  plan-it [+n] <filename                                   *
 *                                                                     *
 *    See "planit.doc" for more details.                               *
 *                                                                     *
 *    This code was developed using the DeSmet C Compiler from C-Ware  *
 *    Corporation, Version 2.2. One brief section uses embedded 8086   *
 *    assembly language to obtain the current date via DOS interrupt   *
 *    21H, function 2AH. Other compilers/operating systems may provide *
 *    better facilities for obtaining this information.                *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define DATLEN 20
#define NMKEY 2
#define NDKEY 7
#define NYKEY 2
#include "stdio.h"

int cy,cm,cd;                 /* Global variables for year, month, day */
long cn;                                               /* century-days */
char cmdy[9];               /* and the string form of the current date */

main (argc,argv)
int argc;
char *argv[];
{
  int today(),show();

  today();                             /* Get today's date from system */
  if (argc==1)
    exit(show(cmdy));                        /* Check today's messages */
  else if (argc==2) {
    exit(show(argv[1]));              /* Check messages for given date */
  }

  puts("\nUsage: plan-it [m/d[/y]]");
  exit(1);
} /* end main */

today()      /* Get the date from DOS and put it into global variables */
{
        long ndays();

#asm
        mov AH,02AH             ;Set AH for DOS function 42, get date
        int 21H                 ;Signal interrupt 21H
        push CX
        pop cy_                 ;Save the year,
        mov cm_,DH              ;  month, and
        mov cd_,DL              ;  day.
#

        cy -= 1900;                      /* Normalize for 20th century */
        sprintf(cmdy,"%d/%d/%d",cm,cd,cy);   /* Convert to string form */
        cn = ndays(cm,cd,cy);               /* Convert to century-days */
        return;
} /* end today */

int mdays[2][13] = {                             /* Days in each month */
  {0,31,28,31,30,31,30,31,31,30,31,30,31},
  {0,31,29,31,30,31,30,31,31,30,31,30,31} };

long sdays(smdy)                /* Convert date string to century-days */
char smdy[];
{
  int i,m,d,y;
  long ndays();

  i = 0;
  m = atoi(smdy);                       /* Get numeric date components */
  while (isdigit(smdy[i]))
    ++i;
  d = atoi(&smdy[++i]);
  while (isdigit(smdy[i]))
    ++i;
  y = atoi(&smdy[++i]);
  if (y == 0)                   /* Use current year if year is missing */
    y = cy;

  return (ndays(m,d,y));
} /* end sdays */

long ndays(m,d,y)              /* Convert numeric date to century-days */
int m,d,y;
{
  int leap,i;

  leap = y%4 == 0;              /* Check for leap year in this century */

  for (i=1; i < m; i++)
    d += mdays[leap][i];

  return (d + y*365 + y/4 - leap);
} /* end ndays */

mdy(n,pm,pd,py)        /* Convert century-days to month, day, and year */
long n;
int *pm,*pd,*py;
{
  int m,d,y,leap;

  y = n/365;
  if (n%365 <= y/4)                              /* Determine the year */
    y--;
  leap = y%4 == 0;                     /* Valid for 20th century years */
  d = n - y*365 - y/4 + leap;                    /* Reduce to year-day */
  for (m=1; d > mdays[leap][m]; m++)            /* Reduce to month-day */
    d -= mdays[leap][m];

  *pm = m;                                       /* Return the results */
  *pd = d;
  *py = y;

  return;
} /* end mdy */

char mkey[2][2][3],             /* Keys for searching the input stream */
     dkey[2][7][4],             /* First subscript toggles between two */
     ykey[2][2][3];             /* sets of keys                        */

show(smdy)                /* Display input messages for the given date */
char smdy[];
{
  int i,j,n,err,leap,scan();
  long ndays();
  char c;

  if (smdy[0] == '+') {            /* Convert "+n" date to date string */
    cn = cn + atoi(&smdy[1]);
    mdy(cn,&cm,&cd,&cy);
    sprintf(smdy,"%d/%d/%d",cm,cd,cy);      /* Produce the date string */
  }

  /* We'll assume we have an unverified date-string, which is not the
     case if we're using the current date or +n methods, but since the
     processing is trivial, we'll keep things simple.                  */

  for (i=j=n=0; (c=smdy[j]) >= '0' && c <= '9' && n < 2; j++,n++)
    ;
  err = c != '/';
  if (!err) {                                       /* Check the month */
    cm = atoi(&smdy[i]);
    err = cm < 1 || cm > 12;
  }

  if (!err) {                              /* Pick off the day portion */
    i = j + 1;
    for (j=i,n=0; (c=smdy[j]) >= '0' && c <= '9' && n < 2; j++,n++)
      ;
    err = c != '/' && c != ' ' && c != '\0';
    if (!err) {
      cd = atoi(&smdy[i]);                /* We'll check the day later */
      if (c == '/') {
        i = j + 1;
        for (j=i; (c=smdy[j]) >= '0' && c <= '9'; j++)
          ;
        err = c != ' ' && c != '\0';
        if (!err)
          cy = atoi(&smdy[i]);
      }
    }

    if (!err) {                             /* Check both year and day */
      err = cy < 0 || cy > 99;
      if (!err) {
        leap = cy%4 == 0;
        err = cd < 1 || cd > mdays[leap][cm];
      }
    }

    if (!err) {
      set_keys(cm,cd,cy,0);                     /* Set the search keys */
      cn = ndays(cm,cd,cy);                    /* Set the century days */
      return (scan());                        /* Scan the input stream */
    }
  }

  printf("planit: invalid date: %s\n",smdy);
  return (err);
} /* end show */

set_keys(m,d,y,n)             /* Define search keys for the given date */
int m,d,y,n;
{
  int lim,leap,month,day,lb;
  static int mcode[2][13] = {                       /* "Magic Numbers" */
    {0,1,4,4,0,2,5,0,3,6,1,4,6},
    {0,0,3,4,0,2,5,0,3,6,1,4,6} };
  static char *weekdays[] = {            /* And the corresponding days */
    "sa","su","m","tu","w","th","f"};

  strcpy(mkey[n][0],"*");                      /* "*" matches any date */
  strcpy(dkey[n][0],"*");
  strcpy(ykey[n][0],"*");

  sprintf(mkey[n][1],"%d",m);              /* Include the numeric date */
  sprintf(dkey[n][1],"%d",d);
  sprintf(ykey[n][1],"%d",y);

  leap = y%4 == 0;                    /* Determine the day of the week */
  month = mcode[leap][m];
  day = (y + y/4 + d + month) % 7;
  strcpy(dkey[n][2],weekdays[day]);

  lim = mdays[leap][m];
  strcpy(dkey[n][3],d==lim ? "l" : "");      /* Last day of the month? */

  lb = day == 6 && d >= lim-2;
  if (!lb)
    lb = day > 1 && d == lim;
  strcpy(dkey[n][4],lb ? "lb" : "");             /* Last business day? */

  sprintf(dkey[n][5],"%d%s",(d+6)/7,dkey[n][2]);   /* nth such weekday */

  if (d+7 > lim)                                 /* Last such weekday? */
    sprintf(dkey[n][6],"l%s",weekdays[day]);
  else
    strcpy(dkey[n][6],"");

  return;
} /* end set_keys */

int scan()        /* Search the input stream using keys from set_keys. */
{
  int i,j,c,incr,lead,match,echo,newline,nkey,m,d,y;
  char datbuf[DATLEN],mm[3],ddd[4],yy[3];
  long sdays();

  match = echo = i = 0;
  newline = 1;

  while ((c = getc(stdin)) != EOF) {              /* Until end of file */
    if (c == '@' && newline) {
      for (i=0; i <= DATLEN; i++)             /* Clear the date buffer */
        datbuf[i] = ' ';
      echo = i = 0;
    }
    else if (!echo) {
      if (i > DATLEN) {
        printf("\nplanit: invalid date in file: %s\n",datbuf);
        return (1);
      }
      else if (c != ' ')
        datbuf[i++] = c;                 /* Accumulate the date string */
      else {
        echo = 1;                             /* Parse the date string */
        lead = incr = nkey = match = 0;

        for (i=0,j=0; (c=datbuf[i]) != '/' && j < 3; i++)
          mm[j++] = c;                                        /* Month */
        mm[j] = '\0';

        for (i++,j=0; (c=datbuf[i]) != '/' && j < 4; i++)
          ddd[j++] = c;                                         /* Day */
        ddd[j] = '\0';

        for (i++,j=0; (c=datbuf[i]) != ' ' &&
                      c != '+' && c != '-' && j < 3; i++)
          yy[j++] = c;
        yy[j] = '\0';                                          /* Year */

        if (c == '+')                                     /* Increment */
          incr = atoi(&datbuf[i+1]);
        else if (c == '-') {                              /* Lead date */
          lead = atoi(&datbuf[i+1]);
          if (lead) {
            mdy(cn+lead,&m,&d,&y);              /* Get the target date */
            nkey = 1;                       /* Point to alternate keys */
            set_keys(m,d,y,nkey);     /* Set the alternate search keys */
          }
        }

        i = 0;
        do {                                           /* Match month? */
          match = strcmp(mkey[nkey][i++],mm) == 0;
        } while (!match && i < NMKEY);

        if (match) {
          i = 0;
          do {                                           /* Match day? */
            match = strcmp(dkey[nkey][i++],ddd) == 0;
          } while (!match && i < NDKEY);

          if (match) {
            i = 0;
            do {                                        /* Match year? */
              match = strcmp(ykey[nkey][i++],yy) == 0;
            } while (!match && i < NYKEY);
          }
        }

        if (!match && incr)                    /* Match periodic date? */
          match = (cn - sdays(datbuf))%incr == 0;

      }                                         /* End of date parsing */
    }
    else if (match) putchar(c);              /* Echo message to screen */
    newline = c == '\n';
  }                                            /* End of scanning loop */
  return (0);
} /* end scan */
